#![warn(missing_docs)]

//! This subcrate is responsible for loading, parsing, validating game content,
//! which is usually stored in `./content`.

mod engine;
mod gun;
mod ship;
mod system;
mod texture;
mod util;

pub use engine::Engine;
pub use gun::{Gun, Projectile};
pub use ship::{EnginePoint, GunPoint, Ship};
pub use system::{Object, System};
pub use texture::Texture;

use anyhow::{bail, Context, Result};
use std::{
	cmp::Eq,
	collections::HashMap,
	fs::File,
	hash::Hash,
	io::Read,
	path::{Path, PathBuf},
};
use toml;
use walkdir::WalkDir;

mod syntax {
	use super::{bail, HashMap, Result};
	use super::{engine, gun, ship, system, texture};
	use serde::Deserialize;

	#[derive(Debug, Deserialize)]
	pub struct Root {
		pub gun: Option<HashMap<String, gun::syntax::Gun>>,
		pub ship: Option<HashMap<String, ship::syntax::Ship>>,
		pub system: Option<HashMap<String, system::syntax::System>>,
		pub engine: Option<HashMap<String, engine::syntax::Engine>>,
		pub texture: Option<HashMap<String, texture::syntax::Texture>>,
	}

	impl Root {
		pub fn new() -> Self {
			Self {
				gun: None,
				ship: None,
				system: None,
				engine: None,
				texture: None,
			}
		}

		pub fn merge(&mut self, other: Root) -> Result<()> {
			// Insert if not exists
			// TODO: replace with a macro and try_insert once that is stable
			if let Some(a) = other.gun {
				if self.gun.is_none() {
					self.gun = Some(a);
				} else {
					let sg = self.gun.as_mut().unwrap();
					for (k, v) in a {
						if sg.contains_key(&k) {
							bail!("Repeated gun name {k}");
						} else {
							sg.insert(k, v);
						}
					}
				}
			}

			if let Some(a) = other.ship {
				if self.ship.is_none() {
					self.ship = Some(a);
				} else {
					let sg = self.ship.as_mut().unwrap();
					for (k, v) in a {
						if sg.contains_key(&k) {
							bail!("Repeated ship name {k}");
						} else {
							sg.insert(k, v);
						}
					}
				}
			}

			if let Some(a) = other.system {
				if self.system.is_none() {
					self.system = Some(a);
				} else {
					let sg = self.system.as_mut().unwrap();
					for (k, v) in a {
						if sg.contains_key(&k) {
							bail!("Repeated system name {k}");
						} else {
							sg.insert(k, v);
						}
					}
				}
			}

			if let Some(a) = other.engine {
				if self.engine.is_none() {
					self.engine = Some(a);
				} else {
					let sg = self.engine.as_mut().unwrap();
					for (k, v) in a {
						if sg.contains_key(&k) {
							bail!("Repeated engine name {k}");
						} else {
							sg.insert(k, v);
						}
					}
				}
			}

			if let Some(a) = other.texture {
				if self.texture.is_none() {
					self.texture = Some(a);
				} else {
					let sg = self.texture.as_mut().unwrap();
					for (k, v) in a {
						if sg.contains_key(&k) {
							bail!("Repeated texture name {k}");
						} else {
							sg.insert(k, v);
						}
					}
				}
			}

			return Ok(());
		}
	}
}

trait Build {
	type InputSyntax;

	/// Build a processed System struct from raw serde data
	fn build(root: Self::InputSyntax, ct: &mut Content) -> Result<()>
	where
		Self: Sized;
}

/// Represents a specific texture defined in the content dir.
#[derive(Debug, Clone, Copy)]
pub struct TextureHandle {
	/// The index of this texture in content.textures
	pub index: usize,

	/// The aspect ratio of this texture (width / height)
	pub aspect: f32,
}

impl Hash for TextureHandle {
	fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
		self.index.hash(state)
	}
}

impl Eq for TextureHandle {}
impl PartialEq for TextureHandle {
	fn eq(&self, other: &Self) -> bool {
		self.index.eq(&other.index)
	}
}

/// Represents generic game content, not connected to any game objects.
#[derive(Debug)]
pub struct Content {
	/// Star systems
	pub systems: Vec<system::System>,

	/// Ship bodies
	pub ships: Vec<ship::Ship>,

	/// Gun outfits
	pub guns: Vec<gun::Gun>,

	/// Engine outfits
	pub engines: Vec<engine::Engine>,

	/// Textures
	pub textures: Vec<texture::Texture>,

	/// Map strings to texture handles
	/// This is never used outside this crate.
	texture_index: HashMap<String, TextureHandle>,

	/// The texture to use for starfield stars
	starfield_handle: Option<TextureHandle>,

	/// Root directory for textures
	texture_root: PathBuf,

	/// Name of starfield texture
	starfield_texture_name: String,
}

impl Content {
	fn try_parse(path: &Path) -> Result<syntax::Root> {
		let mut file_string = String::new();
		let _ = File::open(path)?.read_to_string(&mut file_string);
		let file_string = file_string.trim();
		return Ok(toml::from_str(&file_string)?);
	}

	/// Get the texture handle for the starfield texture
	pub fn get_starfield_handle(&self) -> TextureHandle {
		match self.starfield_handle {
			Some(h) => h,
			None => unreachable!("Starfield texture hasn't been loaded yet!"),
		}
	}

	/// Get a texture from a handle
	pub fn get_texture(&self, h: TextureHandle) -> &texture::Texture {
		// In theory, this could fail if h has a bad index, but that shouldn't ever happen.
		// The only TextureHandles that exist should be created by this crate.
		return &self.textures[h.index];
	}

	/// Load content from a directory.
	pub fn load_dir(
		path: PathBuf,
		texture_root: PathBuf,
		starfield_texture_name: String,
	) -> Result<Self> {
		let mut root = syntax::Root::new();

		for e in WalkDir::new(path).into_iter().filter_map(|e| e.ok()) {
			if e.metadata().unwrap().is_file() {
				// TODO: better warnings
				match e.path().extension() {
					Some(t) => {
						if t.to_str() != Some("toml") {
							println!("[WARNING] {e:#?} is not a toml file, skipping.");
							continue;
						}
					}
					None => {
						println!("[WARNING] {e:#?} is not a toml file, skipping.");
						continue;
					}
				}

				let path = e.path();
				let this_root = Self::try_parse(path)
					.with_context(|| format!("Could not read {}", path.display()))?;

				root.merge(this_root)
					.with_context(|| format!("Could not parse {}", path.display()))?;
			}
		}

		let mut content = Self {
			systems: Vec::new(),
			ships: Vec::new(),
			guns: Vec::new(),
			engines: Vec::new(),
			textures: Vec::new(),
			texture_index: HashMap::new(),
			starfield_handle: None,
			texture_root,
			starfield_texture_name,
		};

		// Order here matters, usually
		if root.texture.is_some() {
			texture::Texture::build(root.texture.take().unwrap(), &mut content)?;
		}
		if root.ship.is_some() {
			ship::Ship::build(root.ship.take().unwrap(), &mut content)?;
		}
		if root.gun.is_some() {
			gun::Gun::build(root.gun.take().unwrap(), &mut content)?;
		}
		if root.engine.is_some() {
			engine::Engine::build(root.engine.take().unwrap(), &mut content)?;
		}
		if root.system.is_some() {
			system::System::build(root.system.take().unwrap(), &mut content)?;
		}

		return Ok(content);
	}
}
